/* $Id: vircam_imcombine.c,v 1.31 2013-10-15 16:40:57 jim Exp $
 *
 * This file is part of the VIRCAM Pipeline
 * Copyright (C) 2005 Cambridge Astronomy Survey Unit
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: jim $
 * $Date: 2013-10-15 16:40:57 $
 * $Revision: 1.31 $
 * $Name: not supported by cvs2svn $
 */

/* Includes */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <math.h>
#include <cpl.h>

#include "vircam_utils.h"
#include "vircam_pfits.h"
#include "vircam_stats.h"
#include "vircam_fits.h"
#include "vircam_mods.h"

/* Macro definitions */

#define FATAL_ERR(_f,_a) {cpl_msg_error(_f,"%s",_a); tidy(); *status = VIR_FATAL; return(*status);}
#define WARN_ERR(_f,_a) {cpl_msg_error(_f,"%s",_a); tidy(); *status = VIR_WARN; return(VIR_WARN);}
#define INFO_ERR(_f,_a) {cpl_msg_info(_f,"%s",_a);}
#define MEDIANCALC 1
#define MEANCALC 2
#define SZBUF 1024

/* Definition of litestruct which is used to hold useful information about
   each of the input frames and their associated images */

typedef struct {
    vir_fits         *frame;
    float            exptime;
    float            expfudge;
    float            skylevel;
    float            skynoise;
    float            skyfudge;
    float            skyrenorm;
} litestruct;

/* Static Global variables */

static litestruct *fileptrs = NULL;
static float **datas = NULL;
static cpl_binary **masks = NULL;
static float *odata;
static cpl_binary *omask;
static long npts;
static int nf;
static float oskylevel;
static long nx;
static long ny;
static unsigned char *rmask = NULL;
static unsigned char *rplus = NULL;

/* Static subroutine prototypes */

static void skyest(float *data, cpl_binary *bpm, float thresh, float *skymed, 
		   float *skynoise);
static void medcalc(float, float, int);
static void meancalc(float, float, int);
static void xclip_med(float thresh, int scaletype);
static void xclip_mean(float thresh, int scaletype);

static void tidy(void);

/**@{*/

/*---------------------------------------------------------------------------*/
/**
    \ingroup reductionmodules
    \brief Stack images into a mean or median image with rejection

    \par Name:
        vircam_imcombine
    \par Purpose:
        Stack images into a mean or median image with rejection
    \par Description:
        A list of frames is given. The images can be stacked by doing pixel
	by pixel means or medians. Images can be scaled or biassed to a common
	background value before any combination or rejection is done. The
	rejection algorithm has the option of doing an extra cycle which looks
	at the area surrounding a rejected pixel to see if the rejection is
	really justified. The output consists of a mean/median image, a 
	rejection mask and a second rejection mask that only marks positive
	rejections. The latter could be useful if trying to assess the number
	of cosmic ray hits.
    \par Language:
        C
    \param fset        
        Input image list
    \param nfits
        The number of input images
    \param combtype    
        Combination type: 1 == Median, 2 == Mean
    \param scaletype   
        Scaling type: 0 == none, 1 == additive, 2 == multiplicative,
        3 == multiplicative by exp time then additive by flux
    \param xrej        
        Extra rejection cycle flag
    \param thresh
        Rejection threshold in sigma above background
    \param outimage
        Output image
    \param rejmask
        Output rejection mask
    \param rejplus
        Output mask of rejected pixels with positive residual
    \param drs
        A propertylist to be used to write DRS info
    \param status
        An input/output status that is the same as the returned values below
    \retval VIR_OK 
        if everything is ok
    \retval VIR_WARN 
        if fset has zero length 
    \retval VIR_FATAL 
        for all other conditions
    \par QC headers:
        None
    \par DRS headers:
        The following DRS keywords are written to the drs propertylist
	- \b PROV****
	    The provenance keywords
    \author
        Jim Lewis, CASU

 */
/*---------------------------------------------------------------------------*/

extern int vircam_imcombine(vir_fits **fset, int nfits, int combtype, 
			    int scaletype, int xrej, float thresh,
			    cpl_image **outimage, unsigned char **rejmask,
			    unsigned char **rejplus, cpl_propertylist **drs,
			    int *status) {
    int i,k,j;
    const char *ic_fctid = "vircam_imcombine";
    char msg[SZBUF];
    float sumsky,sumsig,texp1,texp2,expfudge,skylevel,skynoise,oskynoise;
    float *dat,*work;
    litestruct *ff;
    cpl_image *im;
    cpl_propertylist *plist_p;

    /* Inherited status */

    *rejmask = NULL;
    *rejplus = NULL;
    *drs = NULL;
    *outimage = NULL;
    if (*status != VIR_OK) 
	return(*status);

    /* Check that there are any files in the first place...*/ 

    nf = nfits;
    if (nf == 0) 
	WARN_ERR(ic_fctid,"No files to combine")

    /* Get some file structures */

    fileptrs = cpl_calloc(nf,sizeof(litestruct));

    /* Get workspace for convenience arrays */

    datas = cpl_malloc(nf*sizeof(float *));
    npts = vircam_getnpts(vircam_fits_get_image(fset[0]));
    for (k = 0; k < nf; k++)
	datas[k] = cpl_malloc(npts*sizeof(float));
    masks = cpl_malloc(nf*sizeof(cpl_binary *));

    /* Get pointers to the data arrays */

    for (k = 0; k < nf; k++) {
	im = vircam_fits_get_image(fset[k]);
	dat = cpl_image_get_data_float(im);
	if (dat == NULL) {
	    snprintf(msg,SZBUF,"Failed to load data from extension %d in %s",
		     vircam_fits_get_nexten(fset[k]),
		     vircam_fits_get_filename(fset[k]));
	    FATAL_ERR(ic_fctid,msg)
	}
	for (i = 0; i < npts; i++)
	    datas[k][i] = dat[i];
	masks[k] = cpl_mask_get_data(cpl_image_get_bpm(im));
    }

    /* Open each file in turn and fill in the necessary information. Start with
       the file name...*/

    for (i = 0; i < nf; i++) {
        ff = fileptrs + i;
        ff->frame = fset[i];

        /* If this is the first frame, then keep the size of the data
	   array for future reference */

        if (i == 0) {
	    nx = (long)cpl_image_get_size_x(vircam_fits_get_image(fset[0]));
	    ny = (long)cpl_image_get_size_y(vircam_fits_get_image(fset[0]));
	    npts = nx*ny;
        }

        /* Get the header and the exposure time */

        plist_p = vircam_fits_get_phu(ff->frame);
	if (plist_p == NULL) {
	    snprintf(msg,SZBUF,"Failed to load primary property list %s",
		     vircam_fits_get_filename(ff->frame));
	    INFO_ERR(ic_fctid,msg)
	    texp2 = 1.0;
	} else {
	    if (vircam_pfits_get_exptime(plist_p,&texp2) != VIR_OK) {
		texp2 = 1.0;
		snprintf(msg,SZBUF,"Failed to get exposure time from %s",
			 vircam_fits_get_filename(ff->frame));
		INFO_ERR(ic_fctid,msg)
	    }
	}

        /* Set a few properties */

	ff->exptime = texp2;
        texp1 = fileptrs->exptime;
        expfudge = texp1/texp2;
        ff->expfudge = expfudge;

        /* If scaling by relative exposure time, then do it now. NB: This
           isn't necessary for the first file as all the others are scaled
           relative to it */

        if (scaletype == 3 && i > 0) {
	    for (j = 0; j < npts; j++) 
		datas[i][j] *= ff->expfudge;
	}

        /* Get the background estimate and noise */

        skyest(datas[i],masks[i],thresh,&skylevel,&skynoise);
	ff->skylevel = skylevel;
	ff->skynoise = skynoise;
    }

    /* Work out average background and noise. Then create background zeropoint
       or scale factor, depending upon which was requested in the call */

    work = cpl_malloc(nf*sizeof(float));
    for (i = 0; i < nf; i++)
	work[i] = (fileptrs+i)->skylevel;
    sumsky = vircam_med(work,NULL,(long)nf);
    for (i = 0; i < nf; i++)
	work[i] = (fileptrs+i)->skynoise;
    sumsig = vircam_med(work,NULL,(long)nf);
    cpl_free(work);
    switch (scaletype) {
    case 1:
	for (i = 0; i < nf; i++)
	    (fileptrs+i)->skyfudge = sumsky - (fileptrs+i)->skylevel;
	break;
    case 2:
	for (i = 0; i < nf; i++) 
	    (fileptrs+i)->skyfudge = sumsky/(fileptrs+i)->skylevel;
	break;	    
    case 3:
	for (i = 0; i < nf; i++)
	    (fileptrs+i)->skyfudge = sumsky - (fileptrs+i)->skylevel;
	break;
    default:
	for (i = 0; i < nf; i++) 
	    (fileptrs+i)->skyfudge = 0.0;
	break;
    }

    /* Open an output image based on the first frame in the frameset */

    *outimage = cpl_image_new((cpl_size)nx,(cpl_size)ny,CPL_TYPE_FLOAT);
    odata = cpl_image_get_data_float(*outimage);
    omask = cpl_mask_get_data(cpl_image_get_bpm(*outimage));
    if (*outimage == NULL || odata == NULL) 
	FATAL_ERR(ic_fctid,"Couldn't create output image")
    *rejmask = cpl_calloc(npts,sizeof(*rejmask));
    rmask = *rejmask;
    *rejplus = cpl_calloc(npts,sizeof(*rejplus));
    rplus = *rejplus;

    /* Now do the averaging/medianing */

    switch (combtype) {
    case MEDIANCALC:
	medcalc(thresh,sumsig,scaletype);
	break;
    case MEANCALC:
	meancalc(thresh,sumsig,scaletype);
	break;
    }

    /* Do the extra clipping here if you want it */

    if (xrej) {
        
	/* First get sky background and sigma from output data */

        skyest(odata,omask,thresh,&oskylevel,&oskynoise);

        /* Now loop for all the files subtract off the mean frame (suitably
	   scaled and zero pointed depending on what was done in the first 
	   place) */

        for (i = 0; i < nf; i++) {
	    ff = fileptrs + i;
	    ff->skyrenorm = ff->skylevel/oskylevel;
	    switch (scaletype) {
	    case 1:
		for (k = 0; k < npts; k++) 
		    datas[i][k] -= (odata[k] - oskylevel);
		break;
	    case 2:
		for (k = 0; k < npts; k++)
		    datas[i][k] -= (odata[k] - oskylevel)*ff->skyrenorm;
		break;
	    case 3:
		for (k = 0; k < npts; k++) 
		    datas[i][k] -= (odata[k] - oskylevel);
		break;
	    case 0:
		for (k = 0; k < npts; k++) 
		    datas[i][k] -= (odata[k] - oskylevel);
		break;
	    }

	    /* Re-estimate the noise for this image */

            skyest(datas[i],masks[i],thresh,&skylevel,&skynoise);
	    ff->skynoise = skynoise;
	}

        /* Now do the extra clip... */

	switch (combtype) {
	case MEDIANCALC:
            xclip_med(thresh,scaletype);
	    break;
	case MEANCALC:
	    xclip_mean(thresh,scaletype);
	    break;
	}
    }

    /* Write provenance keywords */

    *drs = cpl_propertylist_new();
    vircam_prov(*drs,fset,nfits);

    /* Right, tidy and get out of here */

    tidy();
    return(VIR_OK);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        xclip_med
    \par Purpose:
        Do an extra clipping cycle for median combined data.
    \par Description:
        The median is recalculated by looking at the pixels that have been 
        rejected and making sure that they really deserve to be rejected 
        based on what the noise properties of the area around the pixel.
	A mask of rejected pixels is set along with a mask of pixels rejected
	with positive residuals.
    \par Language:
        C
    \param thresh
        The threshold for clipping in units of background noise
    \param scaletype
        The type of scaling to be done before combination
    \return Nothing
    \author
        Jim Lewis, CASU
*/
/*---------------------------------------------------------------------------*/

static void xclip_med(float thresh, int scaletype) {
    int nf1,nf2,nfm,nrejmax,is_even,k,is_even2,nrej,nremain,nm,nmm,nplus;
    int nminus,nn,j;
    float **work,*dork,value,cliplev;
    long i;
    litestruct *ff;

    /* Get some workspace */

    work = cpl_malloc(3*sizeof(float *));
    for (i = 0; i < 3; i++)
	work[i] = cpl_malloc(nf*sizeof(float));
    dork = cpl_malloc(nf*sizeof(float));

    /* Loop for each input pixel now... */

    for (i = 0; i < npts; i++) {
	if (omask[i])
	    continue;

	/* Scale or shift data */

	nn = 0;
	switch (scaletype) {
	case 0:
	    for (k = 0; k < nf; k++) {
		if (masks[k][i])
		    continue;
		ff = fileptrs + k;
		work[0][nn] = datas[k][i];
		work[1][nn] = ff->skynoise;
		work[2][nn++] = datas[k][i] + odata[i] - oskylevel;
	    }
	    break;
	case 1:
	    for (k = 0; k < nf; k++) {
		if (masks[k][i])
		    continue;
		ff = fileptrs + k;
		work[0][nn] = datas[k][i] + ff->skyfudge;
		work[1][nn] = ff->skynoise;
		work[2][nn++] = datas[k][i] + odata[i] - oskylevel + 
		    ff->skyfudge;
	    }
	    break;
	case 2:
	    for (k = 0; k < nf; k++) {
		if (masks[k][i])
		    continue;
		ff = fileptrs + k;
		work[0][nn] = datas[k][i]*ff->skyfudge;
		work[1][nn] = ff->skynoise*ff->skyfudge;
		work[2][nn++] = (datas[k][i] + odata[i]*ff->skyrenorm - 
				 ff->skylevel)*ff->skyfudge;
	    }
	    break;
	case 3:
	    for (k = 0; k < nf; k++) {
		if (masks[k][i])
		    continue;
		ff = fileptrs + k;
		work[0][nn] = datas[k][i] + ff->skyfudge;
		work[1][nn] = ff->skynoise;
		work[2][nn++] = datas[k][i] + odata[i] - oskylevel + 
		    ff->skyfudge;
	    }
	    break;
	}	

        /* Set up a few useful variables */

        nf1 = nn/2 - 1;
        nf2 = nf1 + 1;
        nfm = (nn + 1)/2 - 1;
        nrejmax = nn/2;
        is_even = !(nn & 1);

        /* Sort and get a first pass median */

	vircam_sort(work,nn,3);
	if (is_even)
	    value = 0.5*(work[0][nf1] + work[0][nf2]);
        else 
	    if (nn < 5) 
		value = work[0][nfm];
	    else 
		value = 0.25*(work[0][nfm-1] + work[0][nfm+1]) + 0.5*work[0][nfm];
	
	/* Do clipping */

        nplus = 0;
	cliplev = value + thresh*work[1][nn-1];
        while (nplus < nrejmax && work[0][nn-nplus-1] > cliplev)
	    nplus++;
	nminus = 0;
	cliplev = value - thresh*work[1][nn-1];
	while ((nplus+nminus) < nrejmax && work[0][nminus] < cliplev)
	    nminus++;
	nrej = nplus + nminus;

        /* If there were any clipped out, the re-estimate the value */

	if (nrej > 0) {
	    nremain = nn - nrej;
	    if (nremain != 0) {
		nm = nremain/2 - 1;
		for (j = 0; j < nremain; j++) 
		    dork[j] = work[2][j+nminus];
		nmm = (nremain + 1)/2 - 1;
		is_even2 = !(nremain & 1);
		vircam_sort(&dork,nm,1);
		if (is_even2) 
		    value = 0.5*(dork[nm] + dork[nm+1]);
		else 
		    if (nremain < 3) 
			value = dork[nmm];
		    else 
			value = 0.5*dork[nmm] + 0.25*(dork[nmm-1] + dork[nmm+1]);
	    }
	
	    /* Store the result away */

            odata[i] = value;
  	    rmask[i] = min(255,nrej);
	    rplus[i] = min(255,nplus);
	} else {
	    rmask[i] = 0;
	    rplus[i] = 0;
	}
    }

    /* Ditch workspace and get out of here */

    for (i = 0; i < 3; i++)
	cpl_free(work[i]);
    cpl_free(work);
    cpl_free(dork);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        xclip_mean
    \par Purpose:
        Do an extra clipping cycle for mean combined data.
    \par Description:
        The mean is recalculated by looking at the pixels that have been 
        rejected and making sure that they really deserve to be rejected 
        based on what the noise properties of the area around the pixel.
	A mask of rejected pixels is set along with a mask of pixels rejected
	with positive residuals.
    \par Language:
        C
    \param thresh
        The threshold for clipping in units of background noise
    \param scaletype
        The type of scaling to be done before combination
    \return Nothing
    \author
        Jim Lewis, CASU
*/
/*---------------------------------------------------------------------------*/

static void xclip_mean(float thresh, int scaletype) {
    int k,nf2,nrej,nplus,kk,krem;
    float *work[3],value,value2,nrejmax,resid,maxresid;
    long i,nn;
    litestruct *ff;
    unsigned char *iflag;

    /* Get some workspace */

    for (i = 0; i < 3; i++)
	work[i] = cpl_malloc(nf*sizeof(float));
    iflag = cpl_malloc(nf*sizeof(unsigned char));

    /* Loop for each input pixel now... */

    nrejmax = nf/2;
    for (i = 0; i < npts; i++) {
	if (omask[i])
	    continue;

	/* Scale or shift data */

	nn = 0;
	switch (scaletype) {
	case 0:
	    for (k = 0; k < nf; k++) {
		if (masks[k][i])
		    continue;
		ff = fileptrs + k;
		work[0][nn] = datas[k][i];
		work[1][nn] = ff->skynoise;
		work[2][nn] = datas[k][i] + odata[i] - oskylevel;
		iflag[nn++] = 0;
	    }
	    break;
	case 1:
	    for (k = 0; k < nf; k++) {
		if (masks[k][i])
		    continue;
		ff = fileptrs + k;
		work[0][nn] = datas[k][i] + ff->skyfudge;
		work[1][nn] = ff->skynoise;
		work[2][nn] = datas[k][i] + odata[i] - oskylevel + ff->skyfudge;
		iflag[nn++] = 0;
	    }
	    break;
	case 2:
	    for (k = 0; k < nf; k++) {
		if (masks[k][i])
		    continue;
		ff = fileptrs + k;
		work[0][nn] = datas[k][i]*ff->skyfudge;
		work[1][nn] = ff->skynoise*ff->skyfudge;
		work[2][nn] = (datas[k][i] + odata[i]*ff->skyrenorm - 
			       ff->skylevel)*ff->skyfudge;
		iflag[nn++] = 0;
	    }
	    break;
	case 3:
	    for (k = 0; k < nf; k++) {
		if (masks[k][i])
		    continue;
		ff = fileptrs + k;
		work[0][nn] = datas[k][i] + ff->skyfudge;
		work[1][nn] = ff->skynoise;
		work[2][nn] = datas[k][i] + odata[i] - oskylevel + ff->skyfudge;
		iflag[nn++] = 0;
	    }
	    break;
	}	

        /* Get a first pass mean */

	value = 0.0;
	for (k = 0; k < nn; k++) 
	    value += work[0][k];
	value /= (float)nn;
	
        /* Enter a rejection loop. Reject pixels one at a time  */

	nplus = 0;
	nrej = 0;
        for (kk = 0; kk < nrejmax; kk++) {
  	    maxresid = -1.0e30;
            krem = -1;
            for (k = 0; k < nn; k++) {
		if (iflag[k] == 1)
		    continue;
		resid = fabs(work[0][k] - value);
		if (resid > thresh*work[1][k]) {
		    if (nf <= 2)
			resid = work[0][k] - value;
		    if (resid > maxresid) {
			krem = k;
			maxresid = resid;
		    }
		}
	    }
	    
	    /* No further rejections */

	    if (krem == -1)
		break;

	    /* Another pixel is rejected. If it's positive count it */

	    iflag[krem] = 1;
	    if ((work[0][krem] - value) > 0.0)
		nplus++;
	    nrej++;

	    /* Re-evaluate the mean */

    	    nf2 = 0;
	    value2 = 0;
	    for (k = 0; k < nn; k++) {
		if (iflag[k] == 0) {
		    value2 += work[0][k];
		    nf2++;
		}
	    }
	    if (nf2 != 0) 
	        value = value2/(float)nf2;
	    else
		break;
	}

        /* If there were any clipped out store the new value */

	if (nrej > 0) 
            odata[i] = value;
	rmask[i] = min(255,nrej);	
	rplus[i] = min(255,nplus);
    }

    /* Ditch workspace and get out of here */

    for (k = 0; k < 3; k++)
	cpl_free(work[k]);
    cpl_free(iflag);
}


/*---------------------------------------------------------------------------*/
/**
    \par Name:
        medcalc
    \par Purpose:
        Calculate the median of an array of values and clip deviant points
    \par Description:
        An input data array is scaled according to the requested scaling
	algorithm. A median is calculated. A clipping threshold is defined
	from the average sky noise and the median is recalculated if any
        of the scaled pixels fall outside this threshold
    \par Language:
        C
    \param thresh
        The threshold for clipping in units of background noise
    \param avskynoise
        The average noise on the sky
    \param scaletype
        The type of scaling to be done before combination
    \return Nothing
    \author
        Jim Lewis, CASU
*/
/*---------------------------------------------------------------------------*/

static void medcalc(float thresh, float avskynoise, int scaletype) {
    int nf1,nf2,nfm,nrejmax,is_even,nrej,nremain,nm,nmm,is_even2,k,nminus;
    int nplus,nn;
    long i;
    float value,cliplev,*work;

    /* Get a workspace */

    work = cpl_malloc(nf*sizeof(*work));

    /* Ok, loop for each pixel... */

    for (i = 0; i < npts; i++) {

	/* Scale or shift data */

	nn = 0;
	switch (scaletype) {
	case 0:
	    for (k = 0; k < nf; k++) {
		if (masks[k][i])
		    continue;
		work[nn++] = datas[k][i];
	    }
	    break;
	case 1:
	    for (k = 0; k < nf; k++) {
		if (masks[k][i])
		    continue;
		work[nn++] = datas[k][i] + (fileptrs+k)->skyfudge;
	    }
	    break;
	case 2:
	    for (k = 0; k < nf; k++) {
		if (masks[k][i])
		    continue;
		work[nn++] = datas[k][i]*(fileptrs+k)->skyfudge;
	    }
	    break;
	case 3:
	    for (k = 0; k < nf; k++) {
		if (masks[k][i])
		    continue;
		work[nn++] = datas[k][i] + (fileptrs+k)->skyfudge;
	    }
	    break;
        }
	
	/* If nothing is available, then flag the output pixel */

	if (nn == 0) {
	    odata[i] = 0;
	    omask[i] = 1;
	    rmask[i] = 0;
	    rplus[i] = 0;
	    continue;
	}
	    
	/* Set up a few useful variables */

	nf1 = nn/2 - 1;
	nf2 = nf1 + 1;
	nfm = (nn + 1)/2 - 1;
	nrejmax = nn/2;
	is_even = !(nn & 1);

	/* Sort data and get the median */

	vircam_sort(&work,nn,1);
	if (is_even)
	    value = 0.5*(work[nf1] + work[nf2]);
        else 
	    if (nn < 5) 
		value = work[nfm];
	    else 
		value = 0.25*(work[nfm-1] + work[nfm+1]) + 0.5*work[nfm];

	/* Enter a rejection loop */

        nplus = 0;
	cliplev = value + thresh*avskynoise;
        while (nplus < nrejmax && work[nn-nplus-1] > cliplev)
	    nplus++;
	nminus = 0;
	cliplev = value - thresh*avskynoise;
	while ((nplus+nminus) < nrejmax && work[nminus] < cliplev)
	    nminus++;
	nrej = nplus + nminus;

        /* If you've clipped any, then recalculate the median...*/

        if (nrej > 0) {
	    nremain = nn - nrej;
	    nm = nremain/2 - 1 + nminus;
	    nmm = (nremain + 1)/2 - 1 + nminus;
	    is_even2 = !(nremain & 1);
	    if (is_even2) 
		value = 0.5*(work[nm] + work[nm+1]);
	    else 
		if (nremain < 3) 
		    value = work[nmm];
	        else 
		    value = 0.5*work[nmm] + 0.25*(work[nmm-1] + work[nmm+1]);
	}

        /* Store the result away */

        odata[i] = value;
	omask[i] = 0;
	rmask[i] = min(255,nrej);
	rplus[i] = min(255,nplus);
    }

    /* Get rid of workspace */

    cpl_free(work);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        meancalc
    \par Purpose:
        Calculate the mean of an array of values and clip deviant points
    \par Description:
        An input data array is scaled according to the requested scaling
	algorithm. A mean is calculated. A clipping threshold is defined
	from the average sky noise and the mean is recalculated if any
        of the scaled pixels fall outside this threshold
    \par Language:
        C
    \param thresh
        The threshold for clipping in units of background noise
    \param avskynoise
        The average noise on the sky
    \param scaletype
        The type of scaling to be done before combination
    \return Nothing
    \author
        Jim Lewis, CASU
*/
/*---------------------------------------------------------------------------*/

static void meancalc(float thresh, float avskynoise, int scaletype) {
    int nf2,k,nrej,nplus,nrejmax,kk,krem;
    long i,nn;
    float *work,value,value2,maxresid,resid,fresid,cliplev;
    unsigned char *iflag;

    /* Get vectors for workspace */

    work = cpl_malloc(nf*sizeof(*work));
    iflag = cpl_malloc(nf*sizeof(unsigned char));

    /* Ok, loop for each pixel... */

    cliplev = thresh*avskynoise;
    for (i = 0; i < npts; i++) {

	/* Scale or shift data */

	nn = 0;
	switch (scaletype) {
	case 0:
	    for (k = 0; k < nf; k++) {
		if (masks[k][i])
		    continue;
		work[nn++] = datas[k][i];
	    }
	    break;
	case 1:
	    for (k = 0; k < nf; k++) {
		if (masks[k][i])
		    continue;
		work[nn++] = datas[k][i] + (fileptrs+k)->skyfudge;
	    }
	    break;
	case 2:
	    for (k = 0; k < nf; k++) {
		if (masks[k][i])
		    continue;
		work[nn++] = datas[k][i]*(fileptrs+k)->skyfudge;
	    }
	    break;
	case 3:
	    for (k = 0; k < nf; k++) {
		if (masks[k][i])
		    continue;
		work[nn++] = datas[k][i] + (fileptrs+k)->skyfudge;
	    }
	    break;
        }
	
	/* If nothing is available, then flag the output pixel */

	if (nn == 0) {
	    odata[i] = 0;
	    omask[i] = 1;
	    rmask[i] = 0;
	    rplus[i] = 0;
	    continue;
	}
	    
	/* Get the mean */
	
	value = 0.0;
	for (k = 0; k < nn; k++) {
	    value += work[k];
	    iflag[k] = 0;
	}
	value /= (float)nn;

        /* Enter a rejection loop  */
	
        nrejmax = nn - 1;
	nplus = 0;
	nf2 = 0;
	for (kk = 0; kk < nrejmax; kk++) {
	    maxresid = -1.0e30;
	    krem = -1;
	    for (k = 0; k < nn; k++) {
		if (iflag[k] == 1) 
		    continue;
		resid = work[k] - value;
		fresid = (float)fabs((double)resid);
		if (fresid > cliplev) {
		    if (nn <= 2)
			fresid = work[k] - value;
		    if (fresid > maxresid) {
			krem = k;
			maxresid = fresid;
		    }
		}
	    }
	    if (krem == -1)
		break;
	    if ((work[krem] - value) > 0.0)
		nplus++;
            value2 = 0.0;
	    iflag[krem] = 1;
   	    nf2 = 0;
            for (k = 0; k < nn; k++) {
	        if (iflag[k] == 0) {
		    value2 += work[k];
		    nf2 += 1;
		}
	    }
	    value = value2/(float)nf2;
	}

        /* If you've clipped any, then recalculate the mean...*/

	nrej = nn - nf2;

        /* Store the result away */

        odata[i] = value;
	omask[i] = 0;
	rmask[i] = min(255,nrej);
	rplus[i] = min(255,nplus);
    }

    /* Get rid of workspace */

    cpl_free(work);
    cpl_free(iflag);
}
	    
/*---------------------------------------------------------------------------*/
/**
    \par Name:
        skyest
    \par Purpose:
        Calculate a background level and noise estimate
    \par Description:
        A background median and noise level is calculated from an input 
        data array. No bad pixel map is included.
    \par Language:
        C
    \param data
        The data array
    \param mask
        A pixel mask
    \param thresh
        The threshold for clipping in units of background noise
    \param skymed
        The returned sky level
    \param skynoise
        The returned sky noise    
    \return Nothing
    \author
        Jim Lewis, CASU
*/
/*---------------------------------------------------------------------------*/

static void skyest(float *data, cpl_binary *mask, float thresh, float *skymed, 
		   float *skynoise) {
    unsigned char *bpm;

    /* Set up the bad pixel mask */

    bpm = (unsigned char *)mask;

    /* Get the stats */

    vircam_qmedsig(data,bpm,npts,thresh,3,-65535.0,65535.0,skymed,
		   skynoise);

}

/*---------------------------------------------------------------------------*/
/**
    \brief  Tidy up allocated space
 */
/*---------------------------------------------------------------------------*/

static void tidy(void) {
    int i;

    /* Free up work space associated with file structures */

    freespace(fileptrs);
    for (i = 0; i < nf; i++)
	freespace(datas[i]);
    freespace(datas);
    freespace(masks);
}

/**@}*/


/*

$Log: not supported by cvs2svn $
Revision 1.30  2012/01/27 12:25:10  jim
Fixed some casts

Revision 1.29  2012/01/15 17:40:09  jim
Minor modifications to take into accout the changes in cpl API for v6

Revision 1.28  2010/07/13 11:16:50  jim
A few changes to deal with compiler whinges

Revision 1.27  2010/06/03 12:15:31  jim
A few mods to get rid of compiler warnings

Revision 1.26  2008/10/13 08:14:23  jim
Will now take object masks into account

Revision 1.25  2008/09/29 11:27:53  jim
Fixed bug in median clipping algorithm. Also fixed a bug that stopped the
scaling from working in some cases

Revision 1.24  2007/05/14 15:26:03  jim
Fixed bug in multiplicative scaling

Revision 1.23  2007/03/29 12:19:39  jim
Little changes to improve documentation

Revision 1.22  2007/03/01 12:42:41  jim
Modified slightly after code checking

Revision 1.21  2006/10/02 13:47:33  jim
Added missing .h file to include list

Revision 1.20  2006/09/08 09:21:37  jim
Fixed bug in median routine

Revision 1.19  2006/08/21 09:08:10  jim
Fixed a few subtle bugs in the median routines

Revision 1.18  2006/07/04 09:19:05  jim
replaced all sprintf statements with snprintf

Revision 1.17  2006/05/08 10:56:14  jim
Fixed bug where exposure time scaling was not happening to the data copy

Revision 1.16  2006/04/20 11:26:45  jim
Fixed so that original data isn't modified in the input vir_fits list.

Revision 1.15  2006/03/23 21:18:48  jim
Minor changes mainly to comment headers

Revision 1.14  2006/03/22 13:58:31  jim
Cosmetic fixes to keep lint happy

Revision 1.13  2006/03/08 14:32:21  jim
Lots of little modifications

Revision 1.12  2006/03/03 14:29:46  jim
Modified definition of vir_fits and channel table

Revision 1.11  2006/03/01 10:31:28  jim
Now uses new vir_fits objects

Revision 1.10  2006/02/22 10:09:09  jim
Added status variable to call

Revision 1.9  2006/01/23 10:30:49  jim
Mainly documentation mods

Revision 1.8  2005/12/14 22:17:33  jim
Updated docs

Revision 1.7  2005/11/29 11:54:19  jim
Modified call in skyest to allow for larger negative numbers as it is
likely that the reset frames will be negative.

Revision 1.6  2005/11/25 09:56:15  jim
Tidied up some more documentation

Revision 1.5  2005/11/08 12:47:44  jim
Made garbage collection a little better

Revision 1.4  2005/11/07 13:15:16  jim
Fixed lots of bugs and added some error checking

Revision 1.3  2005/11/03 13:28:49  jim
All sorts of changes to tighten up error handling

Revision 1.2  2005/10/14 13:17:54  jim
Now spews out a mask of rejected pixels and rejected positive pixels
(indicative of cosmic ray hits)

Revision 1.1.1.1  2005/08/05 08:29:09  jim
Initial import


*/

